# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd

from pylife.materiallaws import WoehlerCurve
import pylife.stress.rainflow


@pd.api.extensions.register_series_accessor('fatigue')
@pd.api.extensions.register_dataframe_accessor('fatigue')
class Fatigue(WoehlerCurve):

    def damage(self, load_hist):
        cycles = self.basquin_cycles(load_hist.amplitude)

        result = load_hist.cycles.divide(cycles, axis=0)
        result.name = 'damage'
        return result

    def security_load(self, load, cycles, allowed_failure_probability):
        allowed_load = self.basquin_load(cycles, allowed_failure_probability)
        return allowed_load / load.sigma_a

    def security_cycles(self, cycles, load, allowed_failure_probability):
        allowed_cycles = self.basquin_cycles(load.sigma_a, allowed_failure_probability)
        return allowed_cycles / cycles



import sys
import os
import shutil
import xml.etree.ElementTree as ET

coverage_home_folder = sys.argv[1]
print(coverage_home_folder)
coverage_backup_path = f'{coverage_home_folder}coverage_report_backup.xml'
print(coverage_backup_path)
coverage_old_xml_path = f'{coverage_home_folder}coverage_report.xml'
print(coverage_old_xml_path)
coverage_new_xml_path = 'coverage_report.xml'
print(coverage_new_xml_path)

current_folder = os.getcwd() 
print("Current Directory", current_folder) 

is_develop = "develop" in current_folder
print(is_develop)

coverage_old_xml = ET.parse(coverage_old_xml_path)
coverage_new_xml = ET.parse(coverage_new_xml_path)

coverage_new = coverage_new_xml.getroot()
coverage_old = coverage_old_xml.getroot()

if coverage_new.attrib['line-rate'] < coverage_old.attrib['line-rate']:
    raise Exception("Test coverage is decreasing! You have to write unit tests to cover your new code lines")
else:
    print("Test coverage is sufficient")
    if is_develop:
        try:
            os.remove(coverage_backup_path)
            shutil.move(coverage_old_xml_path, coverage_backup_path)
            shutil.copyfile(coverage_new_xml_path, coverage_old_xml_path)    
        except IOError:
            print("One of the coverage files is not accessible")


# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd
import pytest
import unittest.mock as mock

from io import StringIO

from pylife.materialdata import woehler

data = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 8.95e+05],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles']).sample(frac=1)

data_no_mixed_horizons = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles']).sample(frac=1)

no_mixed_horizons_finite_expected = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
]), columns=['load', 'cycles'])

no_mixed_horizons_infinite_expected = pd.DataFrame(np.array([
        [3.50e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles'])

data_pure_runout_horizon_and_mixed_horizons = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07]
]), columns=['load', 'cycles']).sample(frac=1)

pure_runout_horizon_and_mixed_horizons_finite_expected = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05]
    ]), columns=['load', 'cycles'])

pure_runout_horizon_and_mixed_horizons_infinite_expected = pd.DataFrame(np.array([
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07]
]), columns=['load', 'cycles'])

pure_runout_horizon_and_mixed_horizons_finite_expected_conservative = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
    ]), columns=['load', 'cycles'])

pure_runout_horizon_and_mixed_horizons_infinite_expected_conservative = pd.DataFrame(np.array([
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07],
        [2.90e+02, 1.00e+07]
]), columns=['load', 'cycles'])

data_no_runouts = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
]), columns=['load', 'cycles']).sample(frac=1)

no_runouts_infinite_expected = data_no_runouts[:0]
no_runouts_finite_expected = data_no_runouts

data_only_runout_levels = pd.DataFrame(np.array([
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 1.00e+07],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 8.95e+05],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles']).sample(frac=1)

only_runout_levels_infinite_expected = data_only_runout_levels
only_runout_levels_finite_expected = data_only_runout_levels[:0]


data_one_runout_load_level = pd.DataFrame(np.array([
    [3.5000e+02, 8.5930e+05],
    [3.5000e+02, 6.5720e+05],
    [3.5000e+02, 5.8650e+05],
    [3.5000e+02, 1.3085e+06],
    [3.5000e+02, 1.4485e+06],
    [3.0000e+02, 1.0000e+07],
    [3.0000e+02, 1.0000e+07],
    [3.0000e+02, 1.0000e+07],
    [3.0000e+02, 1.0000e+07],
    [3.0000e+02, 1.0000e+07],
    [4.5000e+02, 3.2600e+05],
    [4.5000e+02, 2.6280e+05],
    [4.5000e+02, 1.3200e+05],
    [4.5000e+02, 1.9310e+05],
    [4.5000e+02, 2.5670e+05],
    [3.2500e+02, 2.2402e+06],
    [3.2500e+02, 1.1630e+06],
    [3.2500e+02, 1.2132e+06],
    [3.2500e+02, 1.0347e+06],
    [3.2500e+02, 2.2683e+06]
]), columns=['load', 'cycles']).sample(frac=1)


load_sorted = pd.Series(np.array([
        4.50e+02, 4.50e+02, 4.50e+02, 4.50e+02, 4.00e+02, 4.00e+02, 4.00e+02, 4.00e+02, 3.75e+02, 3.75e+02,
        3.75e+02, 3.75e+02, 3.75e+02, 3.75e+02, 3.50e+02, 3.50e+02, 3.50e+02, 3.50e+02, 3.50e+02, 3.50e+02,
        3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02, 3.25e+02,
        3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02, 3.00e+02]), name='load').sort_values()

cycles_sorted = pd.Series(np.array([
        3.40e+04, 5.40e+04, 6.00e+04, 7.60e+04, 5.30e+04, 9.40e+04, 2.07e+05, 2.27e+05, 6.80e+04, 2.34e+05,
        3.96e+05, 5.00e+05, 6.00e+05, 7.09e+05, 1.70e+05, 1.87e+05, 2.20e+05, 2.89e+05, 3.09e+05, 1.00e+07,
        6.75e+05, 7.51e+05, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07,
        8.95e+05, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07, 1.00e+07]), name='cycles').sort_values()

finite_expected = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
]), columns=['load', 'cycles']).sort_values(by='load').reset_index(drop=True)

infinite_expected = pd.DataFrame(np.array([
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 8.95e+05],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles']).sort_values(by='load').reset_index(drop=True)

finite_expected_conservative = pd.DataFrame(np.array([
        [4.50e+02, 3.40e+04],
        [4.50e+02, 5.40e+04],
        [4.50e+02, 6.00e+04],
        [4.50e+02, 7.60e+04],
        [4.00e+02, 5.30e+04],
        [4.00e+02, 9.40e+04],
        [4.00e+02, 2.07e+05],
        [4.00e+02, 2.27e+05],
        [3.75e+02, 6.80e+04],
        [3.75e+02, 2.34e+05],
        [3.75e+02, 3.96e+05],
        [3.75e+02, 5.00e+05],
        [3.75e+02, 6.00e+05],
        [3.75e+02, 7.09e+05],
]), columns=['load', 'cycles']).sort_values(by='load').reset_index(drop=True)

infinite_expected_conservative = pd.DataFrame(np.array([
        [3.50e+02, 1.70e+05],
        [3.50e+02, 1.87e+05],
        [3.50e+02, 2.20e+05],
        [3.50e+02, 2.89e+05],
        [3.50e+02, 3.09e+05],
        [3.50e+02, 1.00e+07],
        [3.25e+02, 6.75e+05],
        [3.25e+02, 7.51e+05],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.25e+02, 1.00e+07],
        [3.00e+02, 8.95e+05],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07],
        [3.00e+02, 1.00e+07]
]), columns=['load', 'cycles']).sort_values(by='load').reset_index(drop=True)


data_01 = pd.DataFrame({"load": np.array([620, 620, 620, 550, 550, 500, 500, 500, 500, 500, 480, 480]), "cycles": np.array(
    [65783, 89552, 115800, 141826, 190443, 293418, 383341, 525438, 967091, 99505992, 99524024, 199563776])})
data_01_no_pure_runout_horizon = data_01[data_01.load > 480]
data_01_two_fractures = pd.DataFrame({"load": np.array([620, 550,480]), "cycles": np.array([65783, 141826,199563776])})
data_01_one_fracture_level = data_01[data_01.load > 550]


def read_data(s, thres=1e6):
    d = pd.read_csv(StringIO(s), sep="\t", comment="#", names=['cycles', 'load'])
    d.N_threshold = thres
    return d


data_neg_res_1 = read_data("""
127700	450
108000	450
124000	450
127000	450
101000	450
199000	400
213600	400
166000	400
146000	400
140600	400
295000	350
264000	350
330000	350
352000	350
438000	350
645600	300
412000	300
772000	300
501200	300
593600	300
1856000	250
1989900	250
2114500	250
1121400	250
1233600	250
10000000	200
5131700	200
3857300	200
9620900	175
7634000	175
10000000	175
10000000	175
10000000	175""", 10000000)

data_neg_res_2 = read_data("""
109000	377.556025
51000	377.556025
75000	377.556025
111000	377.556025
140000	377.556025
128000	362.84605
75000	362.84605
109000	362.84605
83000	362.84605
68000	362.84605
132000	362.84605
253000	333.4261
987000	333.4261
246000	333.4261
224000	333.4261
159000	333.4261
10000000	304.00615
4576000	304.00615
2139000	304.00615
10000000	304.00615
4576000	304.00615
1191000	289.296175
10000000	289.296175
6337000	289.296175
10000000	289.296175
10000000	289.296175""", 10000000)

all_data = [
    data,
    data_neg_res_1,
    data_neg_res_2
]

def sort_fatigue_data(fd):
    return fd.sort_values(by=['load', 'cycles']).reset_index(drop=True)


def test_fatigue_data_simple_properties():
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_series_equal(fd.load.sort_values(), load_sorted)
    pd.testing.assert_series_equal(fd.cycles.sort_values(), cycles_sorted)

    assert fd.num_runouts == 18
    assert fd.num_fractures == 22


@pytest.mark.parametrize("data, finite_zone_expected, infinite_zone_expected", [
    (data,
     finite_expected,
     infinite_expected),
    (data_no_mixed_horizons,
     no_mixed_horizons_finite_expected,
     no_mixed_horizons_infinite_expected),
    (data_pure_runout_horizon_and_mixed_horizons,
     pure_runout_horizon_and_mixed_horizons_finite_expected,
     pure_runout_horizon_and_mixed_horizons_infinite_expected),
    (data_no_runouts,
     no_runouts_finite_expected,
     no_runouts_infinite_expected),
    (data_only_runout_levels,
     only_runout_levels_finite_expected,
     only_runout_levels_infinite_expected)
])
def test_fatigue_data_finite_infinite_zone(data, finite_zone_expected, infinite_zone_expected):
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.finite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(finite_zone_expected))
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.infinite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(infinite_zone_expected))

@pytest.mark.parametrize("data, finite_zone_expected, infinite_zone_expected", [
    (data,
     finite_expected_conservative,
     infinite_expected_conservative),
    (data_no_mixed_horizons,
     no_mixed_horizons_finite_expected,
     no_mixed_horizons_infinite_expected),
    (data_pure_runout_horizon_and_mixed_horizons,
     pure_runout_horizon_and_mixed_horizons_finite_expected_conservative,
     pure_runout_horizon_and_mixed_horizons_infinite_expected_conservative),
    (data_no_runouts,
     no_runouts_finite_expected,
     no_runouts_infinite_expected)
])
def test_fatigue_data_finite_infinite_zone_conservative(data, finite_zone_expected, infinite_zone_expected):
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.conservative_fatigue_limit()
    print(sort_fatigue_data(fd.finite_zone))
    print(sort_fatigue_data(finite_zone_expected))
    print(fd.fatigue_limit)
    print(sort_fatigue_data(fd.infinite_zone))
    print(sort_fatigue_data(infinite_zone_expected))
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.finite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(finite_zone_expected))
    fd = woehler.determine_fractures(data, 1e7).sort_index().fatigue_data.conservative_fatigue_limit()
    pd.testing.assert_frame_equal(sort_fatigue_data(fd.infinite_zone)[['load', 'cycles']],
                                  sort_fatigue_data(infinite_zone_expected))


def test_woehler_fracture_determination_given():
    df = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4]
    })

    expected = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4],
        'fracture': [True, False, True]
    })

    expected_runouts = pd.DataFrame({
        'load': [2],
        'cycles': [1e7],
        'fracture': [False]
    }, index=[1])

    expected_fractures = pd.DataFrame({
        'load': [1, 3],
        'cycles': [1e6, 1e4],
        'fracture': [True, True]
    }, index=[0, 2])

    test = woehler.determine_fractures(df, 1e7).sort_index()
    pd.testing.assert_frame_equal(test, expected)

    fd = test.fatigue_data
    pd.testing.assert_frame_equal(fd.fractures, expected_fractures)
    pd.testing.assert_frame_equal(fd.runouts, expected_runouts)


def test_woehler_fracture_determination_infered():
    df = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4]
    })

    expected = pd.DataFrame({
        'load': [1, 2, 3],
        'cycles': [1e6, 1e7, 1e4],
        'fracture': [True, False, True]
    })

    test = woehler.determine_fractures(df).sort_index()
    pd.testing.assert_frame_equal(test, expected)


@pytest.mark.parametrize("data, fatigue_limit_expected", [
    (data, 362.5),
    (data_no_mixed_horizons, 362.5),
    (data_pure_runout_horizon_and_mixed_horizons, 362.5),
    (data_no_runouts, 0.0),
    (data_only_runout_levels, 362.5)
])
def test_woehler_endur_zones(data, fatigue_limit_expected):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    assert fd.fatigue_limit == fatigue_limit_expected


@pytest.mark.parametrize("data, fatigue_limit_expected", [
    (data, 325.0),
    (data_no_mixed_horizons, 350.0),
    (data_pure_runout_horizon_and_mixed_horizons, 325.0),
    (data_no_runouts, 0.0),
    (data_only_runout_levels, 325.0)
])
def test_woehler_endur_zones_conservative(data, fatigue_limit_expected):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    print(fd.fatigue_limit)
    fd = fd.conservative_fatigue_limit()
    print(fd.fatigue_limit)
    assert fd.fatigue_limit == fatigue_limit_expected


def test_woehler_endure_zones_no_runouts():
    df = data[data.cycles < 1e7]
    fd = woehler.determine_fractures(df, 1e7).fatigue_data
    assert fd.fatigue_limit == 0.0


def test_woehler_elementary():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 7.0,
        'ND': 3e5,
        'TN': 5.3,
        'TS': 1.27,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_initialize_with_determined_fractures():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 7.0,
        'ND': 3e5,
        'TN': 5.3,
        'TS': 1.27,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7)
    wc = woehler.Elementary(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_initialize_with_pandas_dataframe():
    expected = pd.Series({
        'SD': 362.5,
        'k_1': 7.0,
        'ND': 3e5,
        'TN': 5.3,
        'TS': 1.27,
        'failure_probability': 0.5
    }).sort_index()

    wc = woehler.Elementary(data).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_no_runouts():
    expected = pd.Series({
        'SD': 0.0,
        'k_1': 7.0,
        'TN': 5.3,
        'TS': 1.27,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    wc = woehler.Elementary(fd).analyze().sort_index().drop('ND')
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_elementary_only_one_load_level():
    data = pd.DataFrame(np.array([[350.0, 1e7], [350.0, 1e6]]), columns=['load', 'cycles'])
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.raises(ValueError, match=r"Need at least two load levels to do a Wöhler analysis."):
        woehler.Elementary(fd).analyze().sort_index()


def test_woehler_probit():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 6.94,
        'ND': 463000.,
        'TN': 5.26,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Probit(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_one_runout_load_level():
    fd = woehler.determine_fractures(data_one_runout_load_level, 1e7).fatigue_data
    expected = woehler.Elementary(fd).analyze()
    with pytest.warns(UserWarning, match=r"Probit needs at least two runout load levels. Falling back to Elementary."):
        wc = woehler.Probit(fd).analyze()
    pd.testing.assert_series_equal(wc, expected)


@pytest.mark.filterwarnings("error:invalid")
def test_woehler_probit_data01():
    expected = pd.Series({
        'SD': 490,
        'TS': 1.1,
        'k_1': 8.0,
        'ND': 530e3,
        'TN': 3.0,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_01, 1e7).fatigue_data
    pb = woehler.Probit(fd)
    wc = pb.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_probit_no_runouts():
    expected = pd.Series({
        'SD': 0.,
        'TS': 1.27,
        'k_1': 6.94,
        'ND': 4.4e30,
        'TN': 5.26,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    pb = woehler.Probit(fd)
    with pytest.warns(UserWarning):
        wc = pb.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_max_likelihood_inf_limit():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 6.94,
        'ND': 463000.,
        'TN': 5.26,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.MaxLikeInf(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_bic_without_analysis():
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    with pytest.raises(ValueError, match="^.*BIC.*"):
        we.bayesian_information_criterion()


def test_woehler_max_likelihood_inf_limit_no_runouts():
    expected = pd.Series({
        'SD': 0.,
        'TS': 1.19,
        'k_1': 6.94,
        'ND': 4.4e30,
        'TN': 5.26,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    wc = woehler.MaxLikeInf(fd).analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)


def test_woehler_max_likelihood_full_without_fixed_params():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 6.94,
        'ND': 463000.,
        'TN': 4.7,
        'failure_probability': 0.5
    }).sort_index()

    bic = 45.35256860035525

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    wc = we.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    np.testing.assert_almost_equal(we.bayesian_information_criterion(), bic, decimal=2)


def test_woehler_max_likelihood_full_without_fixed_params_no_runouts():
    expected = pd.Series({
        'SD': 0,
        'TS': 1.,
        'k_1': 6.94,
        'ND': 4.4e30,
        'TN': 5.7,
        'failure_probability': 0.5
    }).sort_index()

    bic = np.inf

    fd = woehler.determine_fractures(data_no_runouts, 1e7).fatigue_data
    we = woehler.MaxLikeFull(fd)
    with pytest.warns(UserWarning, match=r"^.*no runouts are present.*" ):
        wc = we.analyze().sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    np.testing.assert_almost_equal(we.bayesian_information_criterion(), bic, decimal=2)


def test_max_likelihood_full_with_fixed_params():
    expected = pd.Series({
        'SD': 335,
        'TS': 1.19,
        'k_1': 8.0,
        'ND': 520000.,
        'TN': 6.0,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = (
        woehler.MaxLikeFull(fd)
        .analyze(fixed_parameters={'TN': 6.0, 'k_1': 8.0})
        .sort_index()
    )
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)
    assert wc['TN'] == 6.0
    assert wc['k_1'] == 8.0


def test_max_likelihood_full_method_with_all_fixed_params():
    """
    Test of woehler curve parameters evaluation with the maximum likelihood method
    """
    fp = {'k_1': 15.7, 'TN': 1.2, 'SD': 280, 'TS': 1.2, 'ND': 10000000}
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    with pytest.raises(AttributeError, match=r'You need to leave at least one parameter empty!'):
        (
            woehler.MaxLikeFull(fd)
            .analyze(fixed_parameters=fp)
        )


@pytest.mark.parametrize("data,no", [(d, i) for i, d in enumerate(all_data)])
def test_max_likelihood_parameter_sign(data, no):
    def _modify_initial_parameters_mock(fd):
        return fd

    load_cycle_limit = 1e6
    if hasattr(data, "N_threshold"):
        load_cycle_limit = data.N_threshold
    fatdat = woehler.determine_fractures(data, load_cycle_limit=load_cycle_limit)
    ml = woehler.MaxLikeFull(fatigue_data=fatdat.fatigue_data)
    wl = ml.analyze()

    print("Data set number {}".format(no))
    print("Woehler parameters: {}".format(wl))

    def assert_positive_or_nan_but_not_zero(x):
        if np.isfinite(x):
            assert x >= 0
            assert not np.isclose(x, 0.0)

    assert_positive_or_nan_but_not_zero(wl['SD'])
    assert_positive_or_nan_but_not_zero(wl['TS'])
    assert_positive_or_nan_but_not_zero(wl['k_1'])
    assert_positive_or_nan_but_not_zero(wl['ND'])
    assert_positive_or_nan_but_not_zero(wl['TN'])


@pytest.mark.parametrize("invalid_data", [data_01_one_fracture_level, data_01_two_fractures])
def test_max_likelihood_min_three_fractures_on_two_load_levels(invalid_data):
    fd = woehler.determine_fractures(invalid_data, 1e7).fatigue_data
    ml = woehler.MaxLikeFull(fatigue_data=fd)
    with pytest.raises(ValueError, match=r"^.*[N|n]eed at least.*" ):
        ml.analyze()


def test_max_likelihood_one_mixed_horizon():
    expected = pd.Series({
        'SD': 489.3,
        'TS': 1.147,
        'k_1': 7.99,
        'ND': 541e3,
        'TN': 2.51,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data_01, 1e7).fatigue_data
    ml = woehler.MaxLikeFull(fatigue_data=fd)
    with pytest.warns(UserWarning, match=r"^.*less than two mixed load levels.*"):
        wc = ml.analyze().sort_index()
    bic = ml.bayesian_information_criterion()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)

@mock.patch('pylife.materialdata.woehler.bayesian.pm')
def test_bayesian_slope_trace(pm):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    bayes = woehler.Bayesian(fd)
    bayes._nsamples = 1000
    bayes._slope_trace()

    formula, data_dict = pm.glm.GLM.from_formula.call_args[0]
    assert formula == 'y ~ x'
    pd.testing.assert_series_equal(data_dict['x'], np.log10(fd.fractures.load))
    np.testing.assert_array_equal(data_dict['y'], np.log10(fd.fractures.cycles.to_numpy()))
    family = pm.glm.GLM.from_formula.call_args[1]['family']  # Consider switch to kwargs property when py3.7 is dropped
    assert family is pm.glm.families.StudentT()

    pm.sample.assert_called_with(1000, target_accept=0.99, random_seed=None, chains=2, tune=1000)


@mock.patch('pylife.materialdata.woehler.bayesian.pm')
def test_bayesian_TN_trace(pm):
    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    bayes = woehler.Bayesian(fd)
    bayes._common_analysis()
    bayes._nsamples = 1000
    bayes._TN_trace()

    pm.HalfNormal.assert_called_with('stdev', sigma=1.3)

    assert pm.Normal.call_count == 2

    expected_mu = 5.294264482012933
    expected_sigma = 0.2621494419382026

    assert pm.Normal.call_args_list[0][0] == ('mu',)
    np.testing.assert_almost_equal(pm.Normal.call_args_list[0][1]['mu'], expected_mu, decimal=9)
    np.testing.assert_almost_equal(pm.Normal.call_args_list[0][1]['sigma'], expected_sigma, decimal=9)

    assert pm.Normal.call_args_list[1][0] == ('y',)
    observed = pm.Normal.call_args_list[1][1]['observed']  # Consider switch to kwargs property when py3.7 is dropped
    np.testing.assert_almost_equal(observed.mean(), expected_mu, decimal=9)
    np.testing.assert_almost_equal(observed.std(), expected_sigma, decimal=9)

    pm.sample.assert_called_with(1000, target_accept=0.99, random_seed=None, chains=3, tune=1000)


@mock.patch('pylife.materialdata.woehler.bayesian.tt')
@mock.patch('pylife.materialdata.woehler.bayesian.pm')
def test_bayesian_SD_TS_trace_mock(pm, tt):
    def check_likelihood(l, var):
        assert var == tt.as_tensor_variable.return_value
        assert isinstance(l.likelihood, woehler.likelihood.Likelihood)
        np.testing.assert_array_equal(l.likelihood._fd, fd)
        return 'foovar'

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    inf_load_mean = fd.infinite_zone.load.mean()
    inf_load_std = fd.infinite_zone.load.std()

    with mock.patch.object(woehler.Bayesian._LogLike, '__call__', autospec=True) as loglike_call:
        loglike_call.side_effect = check_likelihood

        bayes = woehler.Bayesian(fd)
        bayes._nsamples = 1000
        bayes._SD_TS_trace()

    pm.Normal.assert_called_once_with('SD', mu=inf_load_mean, sigma=inf_load_std * 5)
    pm.Lognormal.assert_called_once()
    np.testing.assert_approx_equal(pm.Lognormal.call_args_list[0][1]['mu'], np.log10(1. / 1.1))
    np.testing.assert_approx_equal(pm.Lognormal.call_args_list[0][1]['sigma'], np.log10(0.5))

    tt.as_tensor_variable.assert_called_once_with([pm.Normal.return_value, pm.Lognormal.return_value])

    pm.Potential.assert_called_once_with('likelihood', 'foovar')

    pm.sample.assert_called_with(1000, cores=1,
                                 chains=3,
                                 random_seed=None,
                                 discard_tuned_samples=True,
                                 tune=1000)


@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._SD_TS_trace')
@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._TN_trace')
@mock.patch('pylife.materialdata.woehler.bayesian.Bayesian._slope_trace')
def test_bayesian_mock(_slope_trace, _TN_trace, _SD_TS_trace):
    expected = pd.Series({
        'SD': 100.,
        'TS': 1.12,
        'k_1': 7.0,
        'ND': 1e6,
        'TN': 5.3,
        'failure_probability': 0.5
    }).sort_index()

    expected_slope_trace = {
        'x': np.array([0.0, -8.0, -6.0]),
        'Intercept': np.array([0.0, 19., 21.])
    }

    expected_SD_TS_trace = {
        'SD': np.array([0.0, 150., 50]),
        'TS': np.array([0.0, 1.22, 1.02])
    }

    _slope_trace.__call__().get_values.side_effect = lambda key: expected_slope_trace[key]
    _TN_trace.__call__().get_values.return_value = np.array([0.0, 5.4, 5.2])
    _SD_TS_trace.__call__().get_values.side_effect = lambda key: expected_SD_TS_trace[key]

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Bayesian(fd).analyze(nsamples=10).sort_index()

    pd.testing.assert_series_equal(wc, expected)


@pytest.mark.slow_acceptance
def test_bayesian_full():
    expected = pd.Series({
        'SD': 340.,
        'TS': 1.12,
        'k_1': 7.0,
        'ND': 400000.,
        'TN': 5.3,
        'failure_probability': 0.5
    }).sort_index()

    fd = woehler.determine_fractures(data, 1e7).fatigue_data
    wc = woehler.Bayesian(fd).analyze(random_seed=4223, progressbar=False).sort_index()
    pd.testing.assert_series_equal(wc, expected, rtol=1e-1)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import cython
import numpy as np

from .general import AbstractDetector


class ThreePointDetector(AbstractDetector):
    r"""Classic three point rainflow counting algorithm.

    See the `here <subsection_TP_>`_ in the demo for an example.

    We take three turning points into account to detect closed hysteresis loops.

    * start: the point where the loop is starting from
    * front: the turning point after the start
    * back: the turning point after the front

    A loop is considered closed if following conditions are met:

    * the load difference between front and back is bigger than or
      equal the one between start and front. In other words: if the
      back goes beyond the starting point. For example (A-B-C) and
      (B-C-D) not closed, whereas (C-D-E) is.

    * the loop init has not been a loop front in a prior closed
      loop. For example F would close the loops (D-E-F) but D is
      already front of the closed loop (C-D-E).

    * the load level of the front has already been covered by a prior
      turning point. Otherwise it is considered part of the front
      residuum.

    When a loop is closed it is possible that the loop back also
    closes unclosed loops of the past by acting as loop back for an
    unclosed start/front pair. For example E closes the loop (C-D-E)
    and then also (A-B-E).

    ::

        Load -----------------------------
        |        x B               F x
        --------/-\-----------------/-----
        |      /   \   x D         /
        ------/-----\-/-\---------/-------
        |    /     C x   \       /
        --\-/-------------\-----/---------
        |  x A             \   /
        --------------------\-/-----------
        |                    x E
        ----------------------------------
        |              Time

    .. _subsection_TP: ../demos/rainflow.ipynb#Classic-Three-Point-Counting
    """

    def __init__(self, recorder):
        """Instantiate a ThreePointDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        super().__init__(recorder)

    @cython.locals(
        start=cython.int, front=cython.int, back=cython.int,
        highest_front=cython.int, lowest_front=cython.int,
        start_val=cython.double, front_val=cython.double, back_val=cython.double,
        turns=cython.double[:])
    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : ThreePointDetector
            The ``self`` object so that processing can be chained
        """
        if len(self._residuals) == 0:
            residuals = samples[:1]
            residual_index = [0, 1]
        else:
            residuals = self._residuals[:-1]
            residual_index = [*range(len(residuals))]

        turns_index, turns_values = self._new_turns(samples)

        turns_np = np.concatenate((residuals, turns_values, samples[-1:]))
        turns_index = np.concatenate((self._residual_index, turns_index))

        turns = turns_np

        highest_front = np.argmax(residuals)
        lowest_front = np.argmin(residuals)

        back = residual_index[-1] + 1
        while back < turns.shape[0]:
            if len(residual_index) >= 2:
                start = residual_index[-2]
                front = residual_index[-1]
                start_val, front_val, back_val = turns[start], turns[front], turns[back]

                if front_val > turns[highest_front]:
                    highest_front = front
                elif front_val < turns[lowest_front]:
                    lowest_front = front
                elif (start >= max(lowest_front, highest_front) and
                      np.abs(back_val - front_val) >= np.abs(front_val - start_val)):
                    self._recorder.record_values(start_val, front_val)
                    self._recorder.record_index(turns_index[start], turns_index[front])
                    residual_index.pop()
                    residual_index.pop()
                    continue

            residual_index.append(back)
            back += 1

        self._residuals = turns_np[residual_index]
        self._residual_index = turns_index[residual_index[:-1]]
        self._recorder.report_chunk(len(samples))

        return self



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import pandas as pd
from scipy import signal as sg
from pylife.stress import timesignal as tsig


def test_resample_acc():
    # sine
    omega = 10*2*np.pi  # Hz
    ts_sin = pd.DataFrame(np.sin(
        omega*np.arange(0, 2, 1/1024)), index=np.arange(0, 2, 1/1024))
    expected_sin = ts_sin.describe().drop(['count', 'mean', '50%', '25%', '75%'])
    test_sin = tsig.TimeSignalPrep(ts_sin).resample_acc(int(12*omega)).describe().drop(
        ['count', 'mean', '50%', '25%', '75%'])
    pd.testing.assert_frame_equal(test_sin, expected_sin, rtol=2)
    # white noise
    ts_wn = pd.DataFrame(np.random.randn(129), index=np.linspace(0, 1, 129))
    expected_wn = ts_wn.describe()
    test_wn = tsig.TimeSignalPrep(ts_wn).resample_acc(128).describe()
    pd.testing.assert_frame_equal(test_wn, expected_wn, check_exact=True)
    # SoR
    t = np.arange(0, 20, 1/4096)
    f = np.logspace(np.log10(1), np.log10(10), num=len(t))
    df = pd.DataFrame(data=np.multiply(
        np.interp(f, np.array([1, 10]), np.array([0.5, 5]), left=0, right=0),
        sg.chirp(t, 1, 20, 10, method="logarithmic", phi=-90)),
        index=t, columns=["Sweep"])
    df['sor'] = df['Sweep'] + 0.1*np.random.randn(len(df))
    expected_sor = df.describe().drop(['count', 'mean', '50%', '25%', '75%'])
    test_sor = tsig.TimeSignalPrep(df).resample_acc(
        2048).describe().drop(['count', 'mean', '50%', '25%', '75%'])
    pd.testing.assert_frame_equal(test_sor, expected_sor, rtol=1)


def test_running_stats_filt():
    t = np.linspace(0, 1, 2048+1)
    sin1 = -abs(150*np.sin(2*np.pi*10*t)+152)
    sin2 = 300*np.sin(2*np.pi*10*t)
    wn1 = np.ones(len(t))
    wn1[500] = -250
    wn2 = np.zeros(len(t))
    wn2[0] = 350
    df = pd.DataFrame(np.hstack((
        sin1,
        sin2,
        wn1,
        wn2)), columns=['data'])
    df_sep = pd.DataFrame(np.vstack((
        sin1,
        sin2,
        wn1,
        wn2))).T
    df_sep.columns = ['sin1', 'sin2', 'wn1', 'wn2']

    test_rms = tsig.TimeSignalPrep(df).running_stats_filt(
        col='data', window_length=2049, buffer_overlap=0.0, limit=0.95, method="rms")
    test_rms.columns = ['sin2']
    pd.testing.assert_frame_equal(test_rms, pd.DataFrame(df_sep['sin2']),
                                  check_index_type=False)

    test_max = tsig.TimeSignalPrep(df).running_stats_filt(
        col='data', window_length=2049, buffer_overlap=0.0, limit=0.95, method="max")
    test_max.columns = ['wn2']
    pd.testing.assert_frame_equal(test_max, pd.DataFrame(df_sep['wn2']),
                                  check_index_type=False)

    test_min = tsig.TimeSignalPrep(df).running_stats_filt(
        col='data', window_length=2049, buffer_overlap=0.0, limit=1, method="min")
    test_min.columns = ['sin1']
    pd.testing.assert_frame_equal(test_min, pd.DataFrame(df_sep['sin1']),
                                  check_index_type=False)

    test_abs = tsig.TimeSignalPrep(df).running_stats_filt(
        col='data', window_length=2049, buffer_overlap=0.0, limit=0.1, method="abs")
    pd.testing.assert_frame_equal(test_abs, df,
                                  check_index_type=False)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import sys
import numpy as np
import odbAccess as ODB


class OdbInterface:

    def __init__(self, odbfile):
        self._odb = ODB.openOdb(odbfile)
        self._asm = self._odb.rootAssembly

        self._index_cache = {}

    def instance_names(self):
        return self._asm.instances.keys()

    def step_names(self):
        return self._odb.steps.keys()

    def frame_names(self, step_name):
        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        return [frame.frameId for frame in step.frames]

    def nodes(self, instance_name, node_set_name):
        instance = self._instance_or_rootasm(instance_name)

        if node_set_name == b'':
            nodes = instance.nodes
        elif node_set_name in instance.nodeSets.keys():
            nodes = instance.nodeSets[node_set_name].nodes
        elif node_set_name in self._asm.nodeSets.keys():
            nodes = self._asm.nodeSets[node_set_name].nodes[self._asm.instances.keys().index(instance_name)]
        else:
            raise KeyError(node_set_name)

        node_data = np.empty((len(nodes), 3))
        index = np.empty(len(nodes), dtype=np.int32)

        for i, nd in enumerate(nodes):
            index[i] = nd.label
            node_data[i] = nd.coordinates

        return (index, node_data)

    def connectivity(self, instance_name, element_set_name):
        instance = self._instance_or_rootasm(instance_name)

        if element_set_name == b'':
            elements = instance.elements
        elif element_set_name in instance.elementSets.keys():
            elements = instance.elementSets[element_set_name].elements
        elif element_set_name in self._asm.elementSets.keys():
            elements = self._asm.elementSets[element_set_name].elements[self._asm.instances.keys().index(instance_name)]
        else:
            raise KeyError(element_set_name)

        index = np.empty(len(elements), dtype=np.int)
        connectivity = []
        for i, el in enumerate(elements):
            index[i] = el.label
            connectivity.append(list(el.connectivity))

        return (index, connectivity)

    def node_sets(self, instance_name):
        instance = self._instance_or_rootasm(instance_name)
        return instance.nodeSets.keys()

    def element_sets(self, instance_name):
        instance = self._instance_or_rootasm(instance_name)
        return instance.elementSets.keys()

    def node_set(self, instance_name, node_set_name):
        instance = self._instance_or_rootasm(instance_name)

        try:
            node_set = instance.nodeSets[node_set_name]
        except Exception as e:
            return e

        nodes = node_set.nodes[0] if instance_name == b'' else node_set.nodes

        return np.array([node.label for node in nodes], dtype=np.int32)

    def element_set(self, instance_name, element_set_name):
        instance = self._instance_or_rootasm(instance_name)

        try:
            element_set = instance.elementSets[element_set_name]
        except Exception as e:
            return e

        elements = element_set.elements[0] if instance_name == b'' else element_set.elements

        return np.array([element.label for element in elements], dtype=np.int32)

    def variable_names(self, step_name, frame_num):
        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        try:
            frame = _get_frame(step, frame_num)
        except Exception as e:
            return e

        return frame.fieldOutputs.keys()

    def variable(self, instance_name, step_name, frame_num, variable_name, node_set_name, element_set_name):

        def block_length(block):
            if block.nodeLabels is not None:
                return block.nodeLabels.shape[0]
            if block.elementLabels is not None:
                return block.elementLabels.shape[0]
            return 0

        def index_block_data(block):
            stack = []
            if block.nodeLabels is not None:
                stack.append(block.nodeLabels)
            if block.elementLabels is not None:
                stack.append(block.elementLabels)

            return np.vstack(stack)

        try:
            step = self._odb.steps[step_name]
        except Exception as e:
            return e

        try:
            frame = _get_frame(step, frame_num)
        except Exception as e:
            return e

        try:
            field = frame.fieldOutputs[variable_name]
        except Exception as e:
            return e

        instance = self._asm.instances[instance_name]

        region = None
        if node_set_name != b'':
            if node_set_name in instance.nodeSets.keys():
                node_set = instance.nodeSets[node_set_name]
            elif node_set_name in self._asm.nodeSets.keys():
                node_set = self._asm.nodeSets[str(node_set_name)]
            else:
                raise KeyError(node_set_name)
            region = node_set

        if element_set_name != b'':
            if element_set_name in instance.elementSets.keys():
                element_set = instance.elementSets[element_set_name]
            elif element_set_name in self._asm.elementSets.keys():
                element_set = self._asm.elementSets[str(element_set_name)]
            else:
                raise KeyError(element_set_name)
            region = element_set

        pos = field.locations[0].position
        if pos == ODB.INTEGRATION_POINT:
            pos = ODB.ELEMENT_NODAL

        if pos != "ABQDEFAULT":
            field = field.getSubset(position=pos)
        if region is not None:
            field = field.getSubset(region=region)

        complabels = field.componentLabels if len(field.componentLabels) > 0 else [variable_name]
        blocks = field.bulkDataBlocks

        length = 0
        for block in blocks:
            if block.instance.name != instance_name:
                continue
            length += block_length(block)

        values = np.empty((length, len(complabels)))

        index_dim = 2 if pos == ODB.ELEMENT_NODAL else 1
        index = np.empty((length, index_dim), dtype=np.int32)

        i = 0
        for block in blocks:
            if block.instance.name != instance_name:
                continue

            block_array = block.data
            size = block_array.shape[0]

            index_block = index_block_data(block)

            index[i:i+size, :] = index_block.T
            values[i:i+size, :] = block_array
            i += size

        if pos == ODB.NODAL:
            index_labels = ['node_id']
        elif pos == ODB.WHOLE_ELEMENT or pos == ODB.CENTROID:
            index_labels = ['element_id']
        else:
            index_labels = ['node_id', 'element_id']
        return (complabels, index_labels, index[:i, :], values[:i])

    def _instance_or_rootasm(self, instance_name):
        if instance_name == b'':
            return self._asm
        try:
            return self._asm.instances[instance_name]
        except Exception as e:
            return e


def _get_frame(step, frame_id):
    for frame in step.frames:
        if frame_id == frame.frameId:
            return frame

    raise Exception("Invalid frame id %s", frame_id)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Vishnu Pradeep"
__maintainer__ = "Johannes Mueller"


import numpy as np

from .general import AbstractDetector


class FourPointDetector(AbstractDetector):
    """ Implements four point rainflow counting algorithm

    We take four turning points into account to detect closed hysteresis loops.

    Consider four consecutive peak/valley points say, A, B, C, and D  If B and C are
    contained within A and B, then a cycle is counted from B to C; otherwise no cycle is
    counted.

    i.e, If X ≥ Y AND Z ≥ Y then a cycle exsist FROM = B and TO = C
    where, ranges X = |D–C|, Y = |C–B|, and Z = |B–A|

    ::

        Load -----------------------------
        |        x B               F x
        --------/-\-----------------/-----
        |      /   \   x D         /
        ------/-----\-/-\---------/-------
        |    /     C x   \       /
        --\-/-------------\-----/---------
        |  x A             \   /
        --------------------\-/-----------
        |                    x E
        ----------------------------------
        |              Time

    So, if a cycle exsist from B to C then delete these peaks from the turns array
    and perform next iteration by joining A&D else if no cylce exsists, then B would
    be the next strarting point.
    """

    def __init__(self, recorder):
        """Instantiate a FourPointDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        super().__init__(recorder)
    def process(self, samples):
        """Process a sample chunk.

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : FourPointDetector
            The ``self`` object so that processing can be chained
        """

        if len(self._residuals) == 0:
            residuals = samples[:1]
            residual_index = [0, 1]
        else:
            residuals = self._residuals[:-1]
            residual_index = [*range(len(residuals))]

        turns_index, turns_values = self._new_turns(samples)

        turns_np = np.concatenate((residuals,turns_values, samples[-1:]))
        turns_index = np.concatenate((self._residual_index, turns_index))

        turns = turns_np
        i = 0

        while i+3 < len(turns):
            ds = np.abs(np.diff(turns)[i:i+3])
            if ds.min() == ds[1]:
                self._recorder.record_values(turns[i+1],turns[i+2])
                self._recorder.record_index(turns_index[i+1], turns_index[i+2])
                turns = np.delete(turns, i+1, 0)
                turns = np.delete(turns, i+1, 0)
                turns_index = np.delete(turns_index, i+1, 0)
                turns_index = np.delete(turns_index, i+1, 0)
                i = max(0, i-4)
            else:
                i += 1
        self._residuals = turns
        self._residual_index = turns_index
        self._recorder.report_chunk(len(samples))

        return self
